#!/usr/bin/env ruby
# written by MOROHOSHI Akihiko <moro@nii.ac.jp>
# $Id: optimizer3.rb,v 1.9 2001/07/29 13:21:37 moro Exp $

StartTime = Time.now

require 'context'
require 'context_sub/context_sub.rb'
#require 'context_sub/context_sub_impl.so'
require 'scanner/smlngscanner'
require 'str_io'
require 'smlngparser'
require 'sorted_array'
require 'set'
require 'forwardable'
require 'tagname'

def max(a,b)
    a > b ? a : b
end
def min(a,b)
    a < b ? a : b
end

if ENV['MORO_TEST_WATCH']
    Thread.abort_on_exception = true
end

WATCH_TIMES = 20
def start_watcher_thread(optimizer)

    if ENV['MORO_TEST_WATCH']
    end

    start_time = Time.now
    end_time = EndTime - 10
    remaining_time = end_time - start_time

    watch_interval = remaining_time.to_f / WATCH_TIMES

    if ENV['MORO_TEST_WATCH']
	STDERR.print "watch_interval = #{watch_interval}\n"
    end
    Thread.start {
	Thread.current.priority += 10
	begin
	    1.upto(WATCH_TIMES) do |i|
		next_wakeup_time = start_time + (watch_interval*i)
		t = Time.now
		interval = (next_wakeup_time - t).floor
		if interval > 0
		    if ENV['MORO_TEST_WATCH']
			STDERR.print "watcher sleep #{interval} sec.\n"
		    end
		    sleep interval
		    if ENV['MORO_TEST_WATCH']
			STDERR.print "watcher wakeup (#{i})\n"
		    end
		end
		if optimizer.search_done
		    if ENV['MORO_TEST_WATCH']
			STDERR.print "\n===> watcher quit.\n"
		    end
		    break
		end
		optimizer.adjust_speed(i.to_f / WATCH_TIMES)
	    end
	rescue Exception => e
	    if ENV['MORO_TEST_WATCH'] || ENV['MORO_TEST']
		STDERR.print "@@@@@@@@@@@@@@@@@@@@\n"
		STDERR.print e,"\n"
		#STDERR.print e.backstrace,"\n"
		raise
	    end
	end
	if ENV['MORO_TEST_WATCH']
	    STDERR.print "\n===> watcher thread end.\n"
	end
    }
end

########################################################################
# optimizer
class Node
    def initialize(parent,index,tag,type)
	@parent,@index,@tag,@type = parent,index,tag,type
	if parent
	    @tagstack = parent.tagstack.dup
	    @search_depth = parent.search_depth
	    case type
	    when :OPEN
		@tagstack.push [tag, parent.context, index]
		@context = context_modified_by_tag parent.context, tag
		@str = tag.open_tag_str
		@overhead_estimated =
		    @str.length*2 + 1 + parent.overhead_estimated
		@search_depth += 1
	    when :CLOSE
		@str = ""
		poptag = nil
		until poptag == tag
		    poptag,@context,idx = @tagstack.pop
		    @str += poptag.close_tag_str
		    unless poptag
			raise 'bug! try to close not found tag.'
		    end
		    @search_depth += 1
		end
	    when nil
		@context = parent.context
		@str = ""
		@search_depth += 1
	    else
		raise ArgumentError,
		    "type should be :OPEN or :CLOSE, but: #{type}"
	    end
	    @overhead = parent.overhead + @str.length
	else
	    # root node
	    @context = Context::RootContext
	    @tagstack = []
	    @overhead = 0
	    @str = ''
	    @overhead_estimated = 0
	    @search_depth = 0
	end
	@overhead_estimated ||= parent.overhead_estimated
    end
    attr_reader :str
    attr_reader :context
    attr_reader :parent,:index,:tag,:type
    attr_reader :tagstack
    attr_reader :overhead	# Υ <b>ʤ3</TT>ʤ5
    attr_reader :overhead_estimated # OPENλΤߡĥ
    attr_reader :search_depth	# õڤο

    def remaining_close_tags_length
	if @remaining_close_tags_length.nil?
	    l = 0
	    @tagstack.each do |tag,context,idx|
		l += tag.length+3
	    end
	    @remaining_close_tags_length = l
	end
	@remaining_close_tags_length
    end

    attr_accessor :estimated_cost # A*  f^(p) (Ŭ)
    def <=>(other)
	# Ĥ
	r = self.estimated_cost <=> other.estimated_cost
	if r == 0
	    # 礭Ĥ
	    r = other.index <=> @index
	end
	r
    end

    include Comparable

    def tagstack_bottomtag
	@tagstack[0][0]
    end
    def tagstack_include_tag?(tag)
	@tagstack.find { |t,c,idx| t == tag }
    end
    def tagstack_empty?
	@tagstack.empty?
    end

    def hash
	@hash ||= (@index*32768).hash + @tagstack.hash
    end
    def eql?(other)
	@index == other.index and
	    @tagstack == other.tagstack
    end

    ####################################################################
    #

    # XXX - dirty, but I have no time...
    @@tokens = nil
    def self.tokens=(tokens)
	@@tokens = tokens 
    end
    def _print_path(buf)
	if @parent
	    @parent._print_path(buf)
	    if @index != @parent.index
		buf << @@tokens[@parent.index].text
	    end
	end
	buf << @str
    end
    protected :_print_path
    
    def release_parent
	unless @path
	    @path = []
	    _print_path(@path)
	    @str = @path
	    @parent = nil
	end
    end
end

class OpenNodes
    extend Forwardable
    include Enumerable
    
    def initialize
	@ary = SortedArray.new
	@hash = Hash.new	# node => [node.estimated_cost, ...]
    end

    def_delegators :@ary, :peek_min, :peek_max, :size, :each

    def add(node)
	@ary.add node
	list = (@hash[node] ||= [])
	list << node.estimated_cost
    end

    def get_max
	if node = @ary.get_max
	    _hash_delete node
	end
	node
    end
    def get_min
	if node = @ary.get_min
	    _hash_delete node
	end
	node
    end

    def delete(node)
	if node = @ary.delete(node)
	    _hash_delete node
	end
	node
    end

    def delete_if(&block)
	@ary.delete_if { |node|
	    if block.call(node)
		_hash_delete(node)
		true
	    else
		false
	    end
	}
    end

    def delete_lesser_equivalent_node(node)
	if list = @hash[node] and list.size > 0
	    if list.find { |est_cost| est_cost > node.estimated_cost }
		self.delete_if { |node2|
		    node2.index == node.index and
			node2.tagstack == node.tagstack and
			node2.estimated_cost > node.estimated_cost
		}
	    end
	    true
	else
	    false
	end
    end

    private
    def _hash_delete(node)
	list = @hash[node]
	list.delete node.estimated_cost
	if list.empty?
	    @hash.delete node
	end
    end
end

USE_OpenNodes = true

class SMLNGOptimizer
    include TagName::Constants
    def initialize(text)
	@text = text
	@tokens = nil
	@tokens_size = nil
	@tag_length_orig = nil

	@shortest_path_leaf = nil

	@search_done = false

	@depth_cutoff = nil

	@max_index = 0
    end
    attr_reader :search_done
    attr_reader :tag_length_orig

    def count_steps
	minstep = (@tokens.size-1) / 2

	(@tokens.size-1).downto(0) do |i|
	    token = @tokens[i]
	    next_token = @tokens[i+1]
	    if next_token
		token.estimated_remaining_cost =
		    token.context.distance_min_est_tag(next_token.context,false) +
		    next_token.estimated_remaining_cost
	    else
		token.estimated_remaining_cost =
		    token.context.distance_min_est_tag(Context::RootContext,false)
	    end
	end
	if ENV['MORO_TEST_COST']
	    @tokens.each do |token|
		print "[#{token.estimated_remaining_cost}]#{token.text}"
	    end
	    print "\n\n"
	end
	
	d = 0
	prev_context = Context::RootContext
	@tokens.size.times do |i|
	    token = @tokens[i]
	    context = token.context
	    d += prev_context.distance_min_est_tag(context)
	    prev_context = context
	end
	d += prev_context.distance_min_est_tag(Context::RootContext)

	# @depth_cutoff = minstep+1
	@depth_cutoff = d
	# ͥõξ硢˲򤬤ǽȤ
	# иŪΨ: 1.1 
	#@depth_cutoff = (d * 1.1).floor
    end

    CLEAN_THRESHOLD_INITIAL = nil # 200
    CLEAN_THRESHOLD_INCREMENT = 50
    CLEAN_COEFFICIENT_INITIAL = 0.3
    CLEAN_COEFFICIENT_MIN = 0.1
    CLEAN_COEFFICIENT_INCREMENT_INITIAL = 0 #0.01
    CLEAN_COEFFICIENT_MAX_INITIAL = 0.9

    def adjust_speed(tick) # tick : 0 -- 1
	return if @search_done or !@search_started

	tokens_size = @tokens_size
	return unless tokens_size
	
	current_max_index = @max_index
	scheduled_index = tokens_size * tick
	if tick > 0.1
	    scheduled_index_2 = (tokens_size + 9 * scheduled_index) / 10
	else
	    scheduled_index_2 = scheduled_index
	end
	if current_max_index > scheduled_index_2
	    # ʤǤ (ahead)
	    if @clean_coefficient > CLEAN_COEFFICIENT_MIN
		@clean_coefficient -= 0.01
	    end
	    if current_max_index < (tokens_size + scheduled_index_2)/2
		@clean_coefficient = scheduled_index.to_f / current_max_index
		if @clean_threshold.nil?
		    @clean_threshold = current_max_index -
			CLEAN_THRESHOLD_INCREMENT
		end
	    end

	    if ENV['MORO_TEST_WATCH']
		STDERR.print "@ ahead: "
	    end
	elsif current_max_index < scheduled_index_2
	    # ٤Ƥ (delaying)
	    delay = (scheduled_index_2 - current_max_index).to_f / tokens_size

	    if @clean_threshold.nil?
		@clean_threshold = current_max_index
	    end
	    
	    if @clean_coefficient_max == @clean_coefficient
		@clean_coefficient_max += (1-@clean_coefficient_max)*0.1
		if @clean_coefficient_max > 0.99
		    @clean_coefficient_max = 0.99
		end
	    end

	    @clean_coefficient =
		min(@clean_coefficient_max,
		    @clean_coefficient + @clean_coefficient_increment)
	    @clean_coefficient =
		max(@clean_coefficient,
		    scheduled_index.to_f / current_max_index)
	    @clean_coefficient_increment += 0.01
	    if ENV['MORO_TEST_WATCH']
		STDERR.print "@ delayed: "
	    end
	end
	if ENV['MORO_TEST_WATCH']
	    STDERR.print "#{current_max_index}/#{scheduled_index}(#{scheduled_index_2})/#{tokens_size}"
	    STDERR.print " --- #{@clean_threshold}/#{@clean_coefficient}/#{@clean_coefficient_max}/#{@clean_coefficient_increment}\n"
	end
    end

    def markup_start
	if USE_OpenNodes
	    @open_nodes = OpenNodes.new
	else
	    @open_nodes = SortedArray.new
	end
	@close_nodes = Set.new
	
	rootnode = Node.new(nil,0,nil,nil)
	rootnode.estimated_cost = @tokens[0].estimated_remaining_cost +
	    @tokens[0].context.distance_min_est_tag(Context::RootContext)
	add_open_node rootnode

	# main loop
	last_token_max_cost = nil
	loop_count = 0
	@clean_threshold = CLEAN_THRESHOLD_INITIAL
	@clean_threshold_increment = CLEAN_THRESHOLD_INCREMENT
	@clean_coefficient = CLEAN_COEFFICIENT_INITIAL
	@clean_coefficient_increment = CLEAN_COEFFICIENT_INCREMENT_INITIAL
	@clean_coefficient_max = CLEAN_COEFFICIENT_MAX_INITIAL

	@search_started = true

	start_watcher_thread(self)
	
	while node = @open_nodes.get_min
	    loop_count += 1
	    last_token_max_cost = node.estimated_cost
	    if markup(node, node.index) == :found
		#print "=== found ===\n"
		break
	    end

	    if ENV['MORO_TEST_COST2']
		print "(#{node.index}/#{@tokens.size}/#{node.estimated_cost}) "
		print "[ ",
		    @open_nodes.map { |node| "#{node.index}/#{node.estimated_cost}" }.join(", "),"]\n"
	    end
	    
	    if ENV['MORO_TEST_COST']
		p @open_nodes.map { |node| node.estimated_cost }
	    end

	    bestnode = @open_nodes.peek_min
	    if @clean_threshold and bestnode and
		    bestnode.index > @clean_threshold
		size0 = @open_nodes.size
		index_min = @clean_coefficient*bestnode.index

		delete_if_proc = proc { |n|
		    if r = n.index < index_min
			n.release_parent
		    end
		    r
		}
		@open_nodes.delete_if &delete_if_proc

		size1 = @open_nodes.size
		if ENV['MORO_TEST_CLEAN']
		    STDERR.print "clean: #{bestnode.index}/#{index_min}: #{size0} -> #{size1}"
		end
		if size0 == size1
		    @clean_threshold_increment *= 2
		else
		    s0 = @close_nodes.size
		    @close_nodes.delete_if &delete_if_proc

		    s1 = @close_nodes.size
		    STDERR.print " (#{s0} -> #{s1})" if ENV['MORO_TEST_CLEAN']
		end
		STDERR.print "\n" if ENV['MORO_TEST_CLEAN']

		if @clean_coefficient < @clean_coefficient_max
		    @clean_coefficient += @clean_coefficient_increment
		end
		@clean_threshold += @clean_threshold_increment
	    end

	    if false and @open_nodes.size > 1000
		500.times do
		    @open_nodes.get_max
		end
	    end

	    if ENV['MORO_TEST_COST3']
		print "idx=#{node.index}/maxidx=#{@open_nodes.map{|n|n.index}.max}/#{@tokens.size}/nodes=#{@open_nodes.size}\n"
	    end
	end
	
	if @shortest_path_leaf.nil?
	    return :not_found
	end

	#print "*** found ***\n"

	while (node = @open_nodes.get_min) and
		(node.estimated_cost <= last_token_max_cost)
	    loop_count += 1
	    markup(node,  node.index)
	    if ENV['MORO_TEST_COST']
		p @open_nodes.map { |node| node.estimated_cost }
	    end
	end

	if ENV['MORO_TEST'] || ENV['MORO_TEST_COST']
	    print "search num: #{loop_count}/#{loop_count+@open_nodes.size}\n"
	end
	
# 	until @shortest_path_leaf
# 	    if ENV['MORO_TEST']
# 		print "depth cutoff=#{@depth_cutoff}\n"
# 	    end
# 	    markup(@rootnode,0)
# 	    @depth_cutoff += 4.5
# 	end
    end

    def minimum_overhead
	if @shortest_path_leaf
	    @shortest_path_leaf.overhead
	else
	    @tag_length_orig
	end
    end

    def add_open_node(node)
	if false # node.parent and node.parent.index == 298
	    print_path_0(STDOUT,node,5)
	    print "\n"
	end

	if USE_OpenNodes
	    if @open_nodes.delete_lesser_equivalent_node(node)
		return
	    end
	else
	    if node2 = @open_nodes.find { |n|
		    n.index == node.index and n.tagstack == node.tagstack
		}
		if node2.estimated_cost > node.estimated_cost
		    @open_nodes.delete node2
		else
		    return
		end
	    end
	end

	if node2 = @close_nodes[node]
	    if node2.estimated_cost > node.estimated_cost
		@close_nodes.delete node2
	    else
		return
	    end
	end

	if @max_index < node.index
	    @max_index = node.index
	end
	@open_nodes.add node
    end

    def open_tag(parent,index,tag)
	p = parent
	while p and p.index == index
	    if p.tag == tag
		if p.type == :CLOSE
		    # ĤФΥ򳫤ȤƤ
		    return
		else
		    if tag != 'U'
			# Ʊ2󳫤ȤƤ
			return
		    end
		end
	    end
	    p = p.parent
	end
	node = Node.new(parent,index,tag,:OPEN)

	if node.overhead_estimated > minimum_overhead
	    if ENV['MORO_TEST'].to_i > 1
		print "(#{index}:#{node.overhead_estimated})"
	    end
	    return
	end

	next_token = @tokens[index+1]
	remaining_cost =  if next_token
			      next_token.estimated_remaining_cost
			  else
			      0
			  end
	node.estimated_cost = node.overhead +
	    node.context.distance_min_est_tag(@tokens[index].context) +
	    node.remaining_close_tags_length +
	    remaining_cost
	    #@tokens[index].estimated_remaining_cost

	if ENV['MORO_TEST_COST']
	    print "    #{node.estimated_cost}) "
	    print_path_0(STDOUT,node)
	    print "\n"
	end

	add_open_node(node)
    end

    def close_tag(parent,index,tag)
	p = parent
	# ĤȤϤäȤ¦ΤΤĤʤФʤʤΤǡ
	# open_tagȰۤʤꤳȽɤ
	# ġĤϤ
	if parent.index == index and parent.type == :OPEN
	    return	      # ФΥĤ褦ȤƤ
 	end
	node = Node.new(parent,index,tag,:CLOSE)

	next_token = @tokens[index+1]
	remaining_cost =  if next_token
			      next_token.estimated_remaining_cost
			  else
			      0
			  end
	node.estimated_cost = node.overhead
	if token = @tokens[index]
	    node.estimated_cost +=
		node.context.distance_min_est_tag(token.context)
	end
	node.estimated_cost +=
	    node.remaining_close_tags_length +
	    remaining_cost
	    #@tokens[index].estimated_remaining_cost

	if ENV['MORO_TEST_COST']
	    print "    #{node.estimated_cost}) "
	    print_path_0(STDOUT,node)
	    print "\n"
	end

	add_open_node node
    end

    def markup(parent, index)
	raise "search failed" if parent.nil?
	@close_nodes.add parent

	old_context = parent.context
	token = @tokens[index]

	if ENV['MORO_TEST_COST']
	    print "#{index}:#{parent.estimated_cost}> "
	    print_path_0(STDOUT,parent)
	    print "\n"
	end

	if token.nil?		# leaf
	    leaf = parent
	    if leaf.tagstack_empty?
		#print "--- found ---\n"
		if (@shortest_path_leaf.nil? or
		    @shortest_path_leaf.overhead > leaf.overhead) and
			@tag_length_orig >= leaf.overhead
		    @shortest_path_leaf = leaf
		end
		return :found
	    end
	    close_tag(leaf,index,leaf.tagstack_bottomtag)
	    return
	end

	new_context = token.context
	if old_context == new_context
	    return markup(parent,index+1)
	end

	pl_contest = context_modified_by_tag(old_context,TAG_PL)
	old_distance = old_context.distance(new_context)
	pl_distance  = pl_contest.distance(new_context)
	pl_valuable = old_distance >= pl_distance

	if old_context.strong_emphasis != new_context.strong_emphasis
	    if ! old_context.strong_emphasis
		open_tag(parent,index,TAG_S)
		if parent.tagstack_include_tag?(TAG_PL)
		    close_tag(parent,index,TAG_PL)
		end
	    else
		close_tag(parent,index,TAG_S)
		open_tag(parent,index,TAG_PL) if pl_valuable
	    end
	end
	if old_context.italic != new_context.italic
	    if ! old_context.italic
		open_tag(parent,index,TAG_I)
		if parent.tagstack_include_tag?(TAG_PL)
		    close_tag(parent,index,TAG_PL)
		end
	    else
		close_tag(parent,index,TAG_I)
		open_tag(parent,index,TAG_PL) if pl_valuable
	    end
	end
	if old_context.bold != new_context.bold
	    if ! old_context.bold
		open_tag(parent,index,TAG_B)
		if parent.tagstack_include_tag?(TAG_PL)
		    close_tag(parent,index,TAG_PL)
		end
	    else
		close_tag(parent,index,TAG_B)
		open_tag(parent,index,TAG_PL) if pl_valuable
	    end
	end
	if old_context.typewriter != new_context.typewriter
	    if ! old_context.typewriter
		open_tag(parent,index,TAG_TT)
		if parent.tagstack_include_tag?(TAG_PL)
		    close_tag(parent,index,TAG_PL)
		end
	    else
		close_tag(parent,index,TAG_TT)
		open_tag(parent,index,TAG_PL) if pl_valuable
	    end
	end

	# EM
	if old_context.emphasis != new_context.emphasis
	    if parent.tagstack_include_tag?(TAG_EM)
		close_tag(parent,index,TAG_EM)
	    end
	    open_tag(parent,index,TAG_EM)
	    if ! new_context.emphasis
		open_tag(parent,index,TAG_PL) if pl_valuable
	    end
	    if parent.tagstack_include_tag?(TAG_PL)
		close_tag(parent,index,TAG_PL)
	    end
	end

	#
	if old_context.underline != new_context.underline
	    if old_context.underline < new_context.underline
		open_tag(parent,index,TAG_U)
		if parent.tagstack_include_tag?(TAG_PL)
		    close_tag(parent,index,TAG_PL)
		end
	    else
		close_tag(parent,index,TAG_U)
		open_tag(parent,index,TAG_PL) if pl_valuable
	    end
	end
	#
	if old_context.color != new_context.color
	    old_tag = TagName.colortag(old_context.color)
	    new_tag = TagName.colortag(new_context.color)
	    if ENV['MORO_TEST_MARKUP']
		print "old_color: #{old_tag}\n"
		print "new_color: #{new_tag}\n"
	    end
	    if old_context.color != Context::COLOR_UNKNOWN
		# </old_color>
		close_tag(parent,index,old_tag)
	    end
	    if new_context.color != Context::COLOR_UNKNOWN
		# <new_color>
		open_tag(parent,index,new_tag)
	    end
	end
	#
	if old_context.size != new_context.size
	    old_tag = TagName.sizetag(old_context.size)
	    new_tag = TagName.sizetag(new_context.size)
	    if ENV['MORO_TEST_MARKUP']
		print "old_size: #{old_tag}\n"
		print "new_size: #{new_tag}\n"
	    end
	    if old_context.size != Context::SIZE_UNKNOWN
		# </old_size>
		close_tag(parent,index,old_tag)
	    end
	    if new_context.size != Context::SIZE_UNKNOWN
		# <new_size>
		open_tag(parent,index,new_tag)
	    end
	end
    end

    def print_path_0(out,node,limit=nil)
	if limit
	    limit -= 1
	    return if limit < 0
	end
	
	if parent = node.parent
	    print_path_0(out,parent,limit)
	    if node.index != parent.index
		out.print "*" if ENV['MORO_TEST_TOKEN_STAR']
		out.print @tokens[parent.index].text
	    end
	end
	if ENV['MORO_TEST_COST'] || ENV['MORO_TEST_PATH_COST']
	    out.print "[#{node.remaining_close_tags_length}/#{node.estimated_cost}]"
	end
	out.print node.str
    end

    def print_path(out)
	if @shortest_path_leaf.nil?
	    raise "path not found!\n"
	end
	print_path_0(out,@shortest_path_leaf)
	if @shortest_path_leaf.index < @tokens.size
	    out.print @tokens[-1].text
	end
    end

    def run(out)
	parser = SMLNGParser.new(@text)
	parser.parse
	@tokens = parser.tokens
	@tokens_size = @tokens.size
	@tokens.freeze

	Node.tokens = @tokens
	
	@tag_length_orig = parser.tag_length
	count_steps
	if ENV['MORO_TEST']
	    print_tokens(out)
	    print "input overhead: #{@tag_length_orig}\n"
	end
	if markup_start == :not_found
	    out.print @text
	else
	    @search_done = true
	    print_path(out)
	end
    end

    def print_forcibly(out)
	if @shortest_path_leaf
	    print_path(out)
	else
	    out.print @text
	end
    end

    def print_tokens(out)
	@tokens.each do |token|
	    c = token.context
	    out.print (c.bold ? 'B' : ' ')
	    out.print (c.emphasis ? 'E' : ' ')
	    out.print (c.strong_emphasis ? 'S' : ' ')
	    out.print (c.italic ? 'I' : ' ')
	    out.print (c.typewriter ? 'T' : ' ')
	    out.print (c.underline.to_s)
	    out.print (c.color == Context::COLOR_UNKNOWN ? ' ' : c.color.chr)
	    out.print (c.size == Context::SIZE_UNKNOWN ? ' ' : c.size.to_s)
	    out.print ' '
	    out.print (token.type == :SPACE ? 'S' : 'T')
	    out.print ' '
	    c = token.estimated_remaining_cost
	    c_1 = c.floor
	    c_2 = (c-c_1)*10
	    out.print (format "%04s.%d",c_1,c_2)
	    out.print ' '
	    out.print '"',token.text.gsub("\n", "\\n"),'"'
	    out.print "\n"
	end
    end
end

########################################################################
# main
if __FILE__==$0
    # »()tlimit롣
    # ̵0Ȥ롣
    tlimit_str = ARGV.shift || ENV['TLIMIT']
    if tlimit_str
	tlimit = tlimit_str.to_i
    end
    tlimit ||= 0

    text = STDIN.read

    optimizer = SMLNGOptimizer.new(text)

    if tlimit > 0
	EndTime = StartTime + tlimit

	#start_watcher_thread(optimizer,EndTime - 10)
	
	Thread.start {
	    # ڤˤʤꤽˤʤä顢ʸ򤽤Τޤ޽Ϥ
	    Thread.current.priority += 2
	    sleep (EndTime - Time.now) - 7
	    begin
		Thread.critical = true
		unless optimizer.search_done
		    optimizer.print_forcibly(STDOUT)
		    exit
		end
	    ensure
		Thread.critical = false
	    end
	}
    end
    Thread.start {
	Thread.current.priority += 1
	loop do
	    GC.disable
	    sleep 60
	    GC.enable
	    #GC.start
	    sleep 5
	end
    }

    if false
	buf = StrIO.new
	optimizer.run(buf)
	print buf, "\n"
    else
	optimizer.run(STDOUT)
	#STDOUT.print "\n"
    end
end
